// instPoint-alpha.h

/*
 * Copyright (c) 1996-2004 Barton P. Miller
 * 
 * We provide the Paradyn Parallel Performance Tools (below
 * described as "Paradyn") on an AS IS basis, and do not warrant its
 * validity or performance.  We reserve the right to update, modify,
 * or discontinue this software at any time.  We shall have no
 * obligation to supply such updates or modifications or any other
 * form of support to you.
 * 
 * This license is for research uses.  For such uses, there is no
 * charge. We define "research use" to mean you may freely use it
 * inside your organization for whatever purposes you see fit. But you
 * may not re-distribute Paradyn or parts of Paradyn, in any form
 * source or binary (including derivatives), electronic or otherwise,
 * to any other organization or entity without our permission.
 * 
 * (for other uses, please contact us at paradyn@cs.wisc.edu)
 * 
 * All warranties, including without limitation, any warranty of
 * merchantability or fitness for a particular purpose, are hereby
 * excluded.
 * 
 * By your use of Paradyn, you understand and agree that we (or any
 * other person or entity with proprietary rights in Paradyn) are
 * under no obligation to provide either maintenance services,
 * update services, notices of latent defects, or correction of
 * defects for Paradyn.
 * 
 * Even if advised of the possibility of such damages, under no
 * circumstances shall we (or any other person or entity with
 * proprietary rights in the software licensed hereunder) be liable
 * to you or any third party for direct, indirect, or consequential
 * damages of any character regardless of type of action, including,
 * without limitation, loss of profits, loss of use, loss of good
 * will, or computer failure or malfunction.  You agree to indemnify
 * us (and any other person or entity with proprietary rights in the
 * software licensed hereunder) for any and all liability it may
 * incur to third parties resulting from your use of Paradyn.
 */

#ifndef _INST_POINT_ALPHA_H_
#define _INST_POINT_ALPHA_H_

#include "dyninstAPI/src/symtab.h"


#ifdef BPATCH_LIBRARY
class BPatch_point;
#endif

class instPoint : public instPointBase {
public:
  instPoint(pd_Function *f, const instruction &instr, Address &adr,
            const bool delayOK,instPointType pointType);
  ~instPoint() {  /* TODO */ }

  pd_Function *getCallee() const { 
     return callIndirect ? NULL : instPointBase::getCallee(); 
  }

  instruction originalInstruction;    /* original instruction */
  instruction delaySlotInsn;  /* original instruction */
  instruction aggregateInsn;  /* aggregate insn */
  instruction otherInstruction;
  instruction isDelayedInsn;  
  instruction inDelaySlotInsn;

  bool inDelaySlot;            /* Is the instruction in a delay slot */
  bool isDelayed;		/* is the instruction a delayed instruction */
  bool callIndirect;		/* is it a call whose target is rt computed ? */
  bool callAggregate;		/* calling a func that returns an aggregate
				   we need to reolcate three insns in this case
				   */
  bool isBranchOut;		/* true if this point is a conditional branch, 
				   that may branch out of the function */
  int branchTarget;		/* the original target of the branch */
  bool leaf;			/* true if the procedure is a leaf     */
  int instId;                      /* id of inst in this function */
  int size;                        /* size of multiple instruction sequences */

};

#endif
