// Automatically generated by Paradyn/igen v4.1-112#0
#ifndef dataManager_thread_BASE_H
#define dataManager_thread_BASE_H
#include "pdutil/h/rpcUtil.h"
#include "pdthread/h/thread.h"

#ifdef IGEN_ERR_ASSERT_ON
#ifndef IGEN_ERR_ASSERT
#define IGEN_ERR_ASSERT assert(0)
#endif
#else
#define IGEN_ERR_ASSERT
#endif


 // Errors that can occur internal to igen
#ifndef _IGEN_ERR_DEF
#define _IGEN_ERR_DEF
typedef enum e_IGEN_ERR {
                       igen_no_err=0,
                       igen_encode_err,
                       igen_decode_err,
                       igen_send_err,
                       igen_read_err,
                       igen_request_err,
                       igen_call_err,
                       igen_proto_err
                       }  IGEN_ERR;

#endif
/*
 * dataManager.h - exported services of the Data Mgr thread of paradyn.
 * $Id: dataManager.I,v 1.1.1.1 2004/10/18 19:49:44 legendre Exp $
 */

#ifndef MMTHREAD_H
#define MMTHREAD_H

#include "common/h/List.h"
#include "common/h/Time.h"
#include "pdutil/h/pdSample.h"
#include "dyninstRPC.xdr.h"
#include "visi.xdr.h"
#include "pdutil/h/stringPool.h"
#include "paradyn/src/DMthread/DMinclude.h"
#include "paradyn/src/DMthread/DVbufferpool.h"

typedef enum {Global, Phase, All} SaveRequestType;

typedef enum { Trace, Sample } dataType;

/*
 * error handler call back.
 *
 */
typedef void (*errorHandler)(int errno, char *message);

/*
 * Handler that gets called when new sample data is delivered.
 *
 *   handle - a performance stream handle from createPerformanceStream
 *   mi - a metricInstance returned by enableDataCollection
 *   buckets - array of sampleValue filled by the DM 
 *   first - the first bucket value
 *   count - number of bucket values in buckets
 *    
 */
typedef void (*sampleDataCallbackFunc)(pdvector<dataValueType> *data,
				       u_int num_data_values);

/*
 * Handler that gets called when new trace data is available.
 *   
 *   handle - a performance stream handle from createPerformanceStream
 *   mi - a metricInstance returned by enableDataCollection
 *   time - time of the event.
 *   eventSize - size of the event specific data.
 *   eventData - event specific data.
 *   Note: traceDataCallbackFunc needs to delete memory at timeStPtr
 */
typedef void (*traceDataCallbackFunc)(perfStreamHandle handle, 
				      metricInstanceHandle mi, 
				      timeStamp *timeStPtr, 
				      int eventSize, 
				      void *eventData);

/*
 * union to hold two types of data callback.
 *
 */
union dataCallback {
    sampleDataCallbackFunc 	sample;
    // Note: traceDataCallbackFunc needs to delete memory at timeStPtr
    traceDataCallbackFunc	trace;
};

typedef union dataCallback dataCallback;

/*
 * Handler that gets called when a new metric is defined.
 *
 * handle - a performance stream handle from createPerformanceStream
 * metricInfo - info about metric, client is resp. for deleting this struct
 *
 */
typedef void (*metricInfoCallback)(perfStreamHandle, 
				   const char *name,
				   int style,
				   int aggregate,
				   const char *units,
				   metricHandle handle,
				   dm_MetUnitsType units_type);

/*
 * Handler that gets called when a new resource is defined.
 *
 * handle - a performance stream handle from createPerformanceStream
 * parent		- parent of new resource
 * newResource		- new resource being created
 * name			- name of the new resource, client must delete
 *
 */
typedef void (*resourceInfoCallback)(perfStreamHandle handle, 
				     resourceHandle parent, 
				     resourceHandle newResource, 
				     const char *name,
				     const char *abstraction);

typedef void (*resourceRetireCallback)(perfStreamHandle handle, 
				       resourceHandle uniqueID, 
				       const char *name);

typedef enum { batchStart, batchEnd } batchMode;

typedef void (*resourceBatchModeCallback)(perfStreamHandle, batchMode mode); 

// histFoldCallback must delete memory at widthPtr
typedef void (*histFoldCallback)(perfStreamHandle, timeLength *width,
                                 phaseType);

// initActValCallback must delete memory at initActValPtr
typedef void (*initActValCallback)(perfStreamHandle, 
				   metricHandle mi,
				   pdSample *initActValPtr,
                                   phaseType phaseID);

typedef enum { appRunning, appPaused, appExited } appState;

typedef void (*appStateChangeCallback)(perfStreamHandle, appState state);

// newPhaseCallback must delete memory at beginPtr, endPtr, bucketWidthPtr
typedef void (*newPhaseCallback)(perfStreamHandle handle, 
				 const char *name, 
				 phaseHandle phase,
				 relTimeStamp *beginPtr, 
				 relTimeStamp *endPtr, 
				 timeLength *bucketWidthPtr,
				 bool with_new_pc,
				 bool with_visis); 

typedef void (*forceFlushCallbackFunc)();

/*
 * Handler that gets called when there is a response to a 
 * getPredictedDataCost call 
 *   
 *   m_handle - the metricHandle assoc. with the request
 *   rl_handle - the resourceListHandle assoc. with the request 
 *   cost     - the cost of enabling the metric/focus pair
 *   clientId - the identifier passed by the calling thread
 */
typedef void (*predDataCostCallbackFunc)(u_int clientId,
					 float cost);

/*
 * Handler that gets called when there is a response to an enableData request 
 *   
 * response -   vector of responses to enable request one element for each   
 *		element in the vector passed by client to enableData request
 * request_id - identifier that client thread associated with this call
 *              (passed to DM on enableDataRequest)
 */
typedef void (*enableDataCallbackFunc)(pdvector<metricInstInfo> *response,
				       u_int request_id, u_int last_cb_for_request);

struct controlCallback {
    metricInfoCallback		mFunc;
    resourceInfoCallback	rFunc;
    resourceRetireCallback	retireFunc;
    resourceBatchModeCallback	bFunc;
    // histFoldCallback must delete memory at widthPtr
    histFoldCallback		fFunc;
    // initActValCallback must delete memory at 
    initActValCallback          avFunc;
    appStateChangeCallback	sFunc;
    // newPhaseCallback must delete memory at beginPtr, endPtr, bucketWidthPtr
    newPhaseCallback	        pFunc;
    predDataCostCallbackFunc    cFunc;
    enableDataCallbackFunc	eFunc;
    forceFlushCallbackFunc      flFunc;
};

typedef struct controlCallback controlCallback;



#endif
class T_dataManager  {
public:
typedef enum {
 } class_ids;
struct T_defineDaemon{ 
const char* command;
const char* dir;
const char* login;
const char* name;
const char* machine;
const char* remote_shell;
const char* flavor;
};
typedef struct T_defineDaemon T_defineDaemon;
struct T_addExecutable{ 
const char* machine;
const char* login;
const char* name;
const char* dir;
const pdvector<pdstring>* argv;
};
typedef struct T_addExecutable T_addExecutable;
struct T_attach{ 
const char* machine;
const char* user;
const char* cmd;
const char* pidStr;
const char* daemonName;
int afterAttach;
};
typedef struct T_attach T_attach;
struct T_saveAllData{ 
const char* dirname;
SaveRequestType optionFlag;
};
typedef struct T_saveAllData T_saveAllData;
struct T_createPerformanceStream{ 
dataType var_12;
dataCallback var_13;
controlCallback var_14;
};
typedef struct T_createPerformanceStream T_createPerformanceStream;
struct T_enableDataRequest{ 
perfStreamHandle ps_handle;
perfStreamHandle pt_handle;
pdvector<metric_focus_pair>* request;
u_int request_Id;
phaseType type;
u_int phaseId;
u_int persistent_data;
u_int persistent_collection;
u_int phase_persistent_data;
};
typedef struct T_enableDataRequest T_enableDataRequest;
struct T_enableDataRequest2{ 
perfStreamHandle ps_handle;
pdvector<metricRLType>* request;
u_int request_Id;
phaseType type;
u_int phaseId;
u_int persistent_data;
u_int persistent_collection;
u_int phase_persistent_data;
};
typedef struct T_enableDataRequest2 T_enableDataRequest2;
struct T_disableDataCollection{ 
perfStreamHandle ps_handle;
perfStreamHandle pt_handle;
metricInstanceHandle var_20;
phaseType var_21;
};
typedef struct T_disableDataCollection T_disableDataCollection;
struct T_disableDataAndClearPersistentData{ 
perfStreamHandle ps_handle;
perfStreamHandle pt_handle;
metricInstanceHandle mi_handle;
phaseType p_type;
bool clear_persistent_data;
bool clear_phase_persistent_data;
};
typedef struct T_disableDataAndClearPersistentData T_disableDataAndClearPersistentData;
struct T_magnify{ 
resourceHandle rh;
resourceListHandle rlh;
magnifyType m;
resourceHandle currentSearchPath;
};
typedef struct T_magnify T_magnify;
struct T_constrain{ 
resourceHandle rh;
resourceListHandle rlh;
};
typedef struct T_constrain T_constrain;
struct T_getPredictedDataCost{ 
perfStreamHandle var_40;
metricHandle var_41;
resourceListHandle var_42;
u_int var_43;
};
typedef struct T_getPredictedDataCost T_getPredictedDataCost;
struct T_getBucketWidth{ 
metricInstanceHandle mih;
phaseType ptype;
};
typedef struct T_getBucketWidth T_getBucketWidth;
struct T_setResourceSearchSuppress{ 
resourceHandle res_handle;
bool newValue;
};
typedef struct T_setResourceSearchSuppress T_setResourceSearchSuppress;
struct T_getMetricValue{ 
metricInstanceHandle var_79;
pdSample* retSample;
};
typedef struct T_getMetricValue T_getMetricValue;
struct T_getSampleValues{ 
metricInstanceHandle mi;
pdSample* buckets;
int numberOfBuckets;
int first;
phaseType phase;
};
typedef struct T_getSampleValues T_getSampleValues;
struct T_getArchiveValues{ 
metricInstanceHandle mi;
pdSample* buckets;
int numberOfBuckets;
int first;
phaseHandle phase_id;
};
typedef struct T_getArchiveValues T_getArchiveValues;
struct T_StartPhase{ 
const relTimeStamp* startTimePtr;
const char* name;
bool with_new_pc;
bool with_visis;
};
typedef struct T_StartPhase T_StartPhase;
struct T_newMetricDefined{ 
metricInfoCallback var_98;
perfStreamHandle var_99;
const char* name;
int style;
int aggregate;
const char* units;
metricHandle handle;
dm_MetUnitsType units_type;
};
typedef struct T_newMetricDefined T_newMetricDefined;
struct T_newResourceDefined{ 
resourceInfoCallback var_101;
perfStreamHandle var_102;
resourceHandle parent;
resourceHandle newResource;
const char* name;
const char* abstr;
};
typedef struct T_newResourceDefined T_newResourceDefined;
struct T_retireResource_{ 
resourceRetireCallback var_104;
perfStreamHandle var_105;
resourceHandle uniqueID;
const char* name;
};
typedef struct T_retireResource_ T_retireResource_;
struct T_changeResourceBatchMode{ 
resourceBatchModeCallback var_107;
perfStreamHandle var_108;
batchMode mode;
};
typedef struct T_changeResourceBatchMode T_changeResourceBatchMode;
struct T_newPerfData{ 
sampleDataCallbackFunc sampleFunc;
pdvector<dataValueType>* data;
u_int num_data_values;
};
typedef struct T_newPerfData T_newPerfData;
struct T_newTracePerfData{ 
traceDataCallbackFunc traceFunc;
pdvector<traceDataValueType>* traceData;
u_int num_traceData_values;
};
typedef struct T_newTracePerfData T_newTracePerfData;
struct T_predictedDataCost{ 
predDataCostCallbackFunc func;
metricHandle m_handle;
resourceListHandle rl_handle;
float cost;
u_int clientId;
};
typedef struct T_predictedDataCost T_predictedDataCost;
struct T_enableDataCallback{ 
enableDataCallbackFunc func;
pdvector<metricInstInfo>* response;
u_int request_id;
u_int last_cb_for_request;
};
typedef struct T_enableDataCallback T_enableDataCallback;
struct T_histFold{ 
histFoldCallback var_115;
perfStreamHandle var_116;
timeLength* widthPtr;
phaseType var_117;
};
typedef struct T_histFold T_histFold;
struct T_setInitialActualValue{ 
initActValCallback var_119;
perfStreamHandle var_120;
metricHandle mi;
pdSample* initActValPtr;
phaseType phaseID;
};
typedef struct T_setInitialActualValue T_setInitialActualValue;
struct T_changeState{ 
appStateChangeCallback var_122;
perfStreamHandle var_123;
appState var_124;
};
typedef struct T_changeState T_changeState;
struct T_newPhaseInfo{ 
newPhaseCallback var_126;
perfStreamHandle var_127;
const char* var_128;
phaseHandle var_129;
relTimeStamp* startPtr;
relTimeStamp* endPtr;
timeLength* bucketWidthPtr;
bool var_130;
bool var_131;
};
typedef struct T_newPhaseInfo T_newPhaseInfo;
enum message_tags {
  verify = 2500,
  defineDaemon_REQ, defineDaemon_RESP,
  addExecutable_REQ, addExecutable_RESP,
  attach_REQ, attach_RESP,
  applicationDefined_REQ, applicationDefined_RESP,
  startApplication_REQ, startApplication_RESP,
  pauseApplication_REQ, pauseApplication_RESP,
  createCallGraph_REQ, createCallGraph_RESP,
  pauseProcess_REQ, pauseProcess_RESP,
  continueApplication_REQ, continueApplication_RESP,
  detachApplication_REQ, detachApplication_RESP,
  saveAllData_REQ, saveAllData_RESP,
  saveAllResources_REQ, saveAllResources_RESP,
  createPerformanceStream_REQ, createPerformanceStream_RESP,
  destroyPerformanceStream_REQ, destroyPerformanceStream_RESP,
  enableDataRequest_REQ, enableDataRequest_RESP,
  enableDataRequest2_REQ, enableDataRequest2_RESP,
  disableDataCollection_REQ, disableDataCollection_RESP,
  disableDataAndClearPersistentData_REQ, disableDataAndClearPersistentData_RESP,
  magnify_REQ, magnify_RESP,
  constrain_REQ, constrain_RESP,
  morespecific_REQ, morespecific_RESP,
  isSuppressed_REQ, isSuppressed_RESP,
  getFocusNameFromMI_REQ, getFocusNameFromMI_RESP,
  getFocusName_REQ, getFocusName_RESP,
  getFocusNameFromHandle_REQ, getFocusNameFromHandle_RESP,
  setPersistentCollection_REQ, setPersistentCollection_RESP,
  clearPersistentCollection_REQ, clearPersistentCollection_RESP,
  setPersistentData_REQ, setPersistentData_RESP,
  clearPersistentData_REQ, clearPersistentData_RESP,
  getPredictedDataCost_REQ, getPredictedDataCost_RESP,
  getBucketWidth_REQ, getBucketWidth_RESP,
  getCurrentBucketWidth_REQ, getCurrentBucketWidth_RESP,
  getGlobalBucketWidth_REQ, getGlobalBucketWidth_RESP,
  getCurrentPhaseId_REQ, getCurrentPhaseId_RESP,
  getCurrentStartTime_REQ, getCurrentStartTime_RESP,
  getMaxBins_REQ, getMaxBins_RESP,
  getRootResource_REQ, getRootResource_RESP,
  getRootResources_REQ, getRootResources_RESP,
  getResourceList_REQ, getResourceList_RESP,
  getResourceHandles_REQ, getResourceHandles_RESP,
  matchMetFocus_REQ, matchMetFocus_RESP,
  findResource_REQ, findResource_RESP,
  getResourceLabelName_REQ, getResourceLabelName_RESP,
  getResourceName_REQ, getResourceName_RESP,
  findResourceList_REQ, findResourceList_RESP,
  setResourceSearchSuppress_REQ, setResourceSearchSuppress_RESP,
  setResourceSearchChildrenSuppress_REQ, setResourceSearchChildrenSuppress_RESP,
  setResourceInstSuppress_REQ, setResourceInstSuppress_RESP,
  getAvailableMetrics_REQ, getAvailableMetrics_RESP,
  getAvailableMetInfo_REQ, getAvailableMetInfo_RESP,
  findMetric_REQ, findMetric_RESP,
  getMetricName_REQ, getMetricName_RESP,
  getMetricInfo_REQ, getMetricInfo_RESP,
  getMetric_REQ, getMetric_RESP,
  getMetricNameFromMI_REQ, getMetricNameFromMI_RESP,
  getMetricValue_REQ, getMetricValue_RESP,
  getSampleValues_REQ, getSampleValues_RESP,
  getArchiveValues_REQ, getArchiveValues_RESP,
  getTotValue_REQ, getTotValue_RESP,
  StartPhase_REQ, StartPhase_RESP,
  getAllPhaseInfo_REQ, getAllPhaseInfo_RESP,
  printResources_REQ, printResources_RESP,
  printStatus_REQ, printStatus_RESP,
  coreProcess_REQ, coreProcess_RESP,
  printDaemons_REQ, printDaemons_RESP,
  displayDaemonStartInfo_REQ, displayDaemonStartInfo_RESP,
  printDaemonStartInfo_REQ, printDaemonStartInfo_RESP,
  displayParadynGeneralInfo_REQ, displayParadynGeneralInfo_RESP,
  displayParadynLicenseInfo_REQ, displayParadynLicenseInfo_RESP,
  displayParadynReleaseInfo_REQ, displayParadynReleaseInfo_RESP,
  displayParadynVersionInfo_REQ, displayParadynVersionInfo_RESP,
  getAvailableDaemons_REQ, getAvailableDaemons_RESP,
  newMetricDefined_REQ, newMetricDefined_RESP,
  newResourceDefined_REQ, newResourceDefined_RESP,
  retireResource__REQ, retireResource__RESP,
  changeResourceBatchMode_REQ, changeResourceBatchMode_RESP,
  newPerfData_REQ, newPerfData_RESP,
  newTracePerfData_REQ, newTracePerfData_RESP,
  predictedDataCost_REQ, predictedDataCost_RESP,
  enableDataCallback_REQ, enableDataCallback_RESP,
  forceFlush_REQ, forceFlush_RESP,
  histFold_REQ, histFold_RESP,
  setInitialActualValue_REQ, setInitialActualValue_RESP,
  changeState_REQ, changeState_RESP,
  newPhaseInfo_REQ, newPhaseInfo_RESP,
  error, last }; 
 typedef enum message_tags message_tags;
};


#endif

